from typing import Dict
import numpy as np


class SparseMask:
    def __init__(self, mask_array:np.ndarray, roi_name_dict: Dict[str, int]):
        self.mask_array = mask_array
        self.roi_name_dict = roi_name_dict



from typing import Tuple, List
import os
import pathlib
import glob
import json
import numpy as np
import matplotlib.pyplot as plt


def markdown_report_images(output_folder, modality_count):
    modalities = list(modality_count.keys())
    modality_totals = list(modality_count.values())
    if not os.path.exists(pathlib.Path(output_folder, "markdown_images").as_posix()):
        os.makedirs(pathlib.Path(output_folder, "markdown_images").as_posix())
    plt.figure(1)
    plt.bar(modalities, modality_totals)
    plt.savefig(pathlib.Path(output_folder, "markdown_images", "nnunet_modality_count.png").as_posix())

    plt.figure(2)
    train_total = len(glob.glob(pathlib.Path(output_folder, "labelsTr", "*.nii.gz").as_posix()))
    test_total = len(glob.glob(pathlib.Path(output_folder, "labelsTs", "*.nii.gz").as_posix()))
    plt.pie([train_total, test_total], labels=[f"Train - {train_total}", f"Test - {test_total}"])
    plt.savefig(pathlib.Path(output_folder, "markdown_images", "nnunet_train_test_pie.png").as_posix())

# this code is taken from:
# Division of Medical Image Computing, German Cancer Research Center (DKFZ)
# in the nnUNet and batchgenerator repositories


def save_json(obj, file: str, indent: int = 4, sort_keys: bool = True) -> None:
    with open(file, 'w') as f:
        json.dump(obj, f, sort_keys=sort_keys, indent=indent)


def get_identifiers_from_splitted_files(folder: str):
    uniques = np.unique([i[:-12] for i in subfiles(folder, suffix='.nii.gz', join=False)])
    return uniques


def subfiles(folder: str, join: bool = True, prefix: str = None, suffix: str = None, sort: bool = True) -> List[str]:
    if join:
        path_fn = os.path.join
    else:
        def path_fn(x, y): return y
        
    res = [path_fn(folder, i) for i in os.listdir(folder) if os.path.isfile(os.path.join(folder, i))
           and (prefix is None or i.startswith(prefix))
           and (suffix is None or i.endswith(suffix))]
    if sort:
        res.sort()
    return res


def generate_dataset_json(output_file: str, imagesTr_dir: str, imagesTs_dir: str, modalities: Tuple,
                          labels: dict, dataset_name: str, sort_keys=True, license: str = "hands off!", dataset_description: str = "",
                          dataset_reference="", dataset_release='0.0'):
    """
    :param output_file: This needs to be the full path to the dataset.json you intend to write, so
    output_file='DATASET_PATH/dataset.json' where the folder DATASET_PATH points to is the one with the
    imagesTr and labelsTr subfolders
    :param imagesTr_dir: path to the imagesTr folder of that dataset
    :param imagesTs_dir: path to the imagesTs folder of that dataset. Can be None
    :param modalities: tuple of strings with modality names. must be in the same order as the images (first entry
    corresponds to _0000.nii.gz, etc). Example: ('T1', 'T2', 'FLAIR').
    :param labels: dict with int->str (key->value) mapping the label IDs to label names. Note that 0 is always
    supposed to be background! Example: {0: 'background', 1: 'edema', 2: 'enhancing tumor'}
    :param dataset_name: The name of the dataset. Can be anything you want
    :param sort_keys: In order to sort or not, the keys in dataset.json
    :param license:
    :param dataset_description:
    :param dataset_reference: website of the dataset, if available
    :param dataset_release:
    :return:
    """
    train_identifiers = get_identifiers_from_splitted_files(imagesTr_dir)

    if imagesTs_dir is not None:
        test_identifiers = get_identifiers_from_splitted_files(imagesTs_dir)
    else:
        test_identifiers = []

    json_dict = {}
    json_dict['name'] = dataset_name
    json_dict['description'] = dataset_description
    json_dict['tensorImageSize'] = "4D"
    json_dict['reference'] = dataset_reference
    json_dict['licence'] = license
    json_dict['release'] = dataset_release
    json_dict['modality'] = {str(i): modalities[i] for i in range(len(modalities))}
    json_dict['labels'] = {str(i): labels[i] for i in labels.keys()}

    json_dict['numTraining'] = len(train_identifiers)
    json_dict['numTest'] = len(test_identifiers)
    json_dict['training'] = [
        {'image': "./imagesTr/%s.nii.gz" % i, "label": "./labelsTr/%s.nii.gz" % i} for i
        in
        train_identifiers]
    json_dict['test'] = ["./imagesTs/%s.nii.gz" % i for i in test_identifiers]

    if not output_file.endswith("dataset.json"):
        print("WARNING: output file name is not dataset.json! This may be intentional or not. You decide. "
              "Proceeding anyways...")
    save_json(json_dict, os.path.join(output_file), sort_keys=sort_keys)



from argparse import ArgumentParser
import os
import pathlib
import glob
import json
import pandas as pd
from pydicom import dcmread
from tqdm import tqdm
from joblib import Parallel, delayed


def crawl_one(folder):
    folder_path = pathlib.Path(folder)
    database = {}
    for path, _, _ in os.walk(folder):
        # find dicoms
        dicoms = glob.glob(pathlib.Path(path, "**", "*.dcm").as_posix(), recursive=True)
        # print('\n', folder, dicoms)
        # instance (slice) information
        for dcm in dicoms:
            try:
                dcm_path  = pathlib.Path(dcm)
                # parent    = dcm_path.parent#.as_posix()
                fname     = dcm_path.name
                rel_path  = dcm_path.relative_to(folder_path.parent.parent)                                        # rel_path of dicom from folder
                rel_posix = rel_path.parent.as_posix()     # folder name + until parent folder of dicom

                meta      = dcmread(dcm, force=True, stop_before_pixels=True)
                patient   = str(meta.PatientID)
                study     = str(meta.StudyInstanceUID)
                series    = str(meta.SeriesInstanceUID)
                instance  = str(meta.SOPInstanceUID)

                reference_ct, reference_rs, reference_pl,  = "", "", ""
                tr, te, tesla, scan_seq, elem = "", "", "", "", ""
                try:
                    orientation = str(meta.ImageOrientationPatient)  # (0020, 0037)
                except:
                    orientation = ""

                try:
                    orientation_type = str(meta.AnatomicalOrientationType)  # (0010, 2210)
                except:
                    orientation_type = ""

                try:  # RTSTRUCT
                    reference_ct = str(meta.ReferencedFrameOfReferenceSequence[0].RTReferencedStudySequence[0].RTReferencedSeriesSequence[0].SeriesInstanceUID)
                except: 
                    try: # SEGMENTATION
                        reference_ct = str(meta.ReferencedSeriesSequence[0].SeriesInstanceUID)
                    except:
                        try:  # RTDOSE
                            reference_rs = str(meta.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID)
                        except:
                            pass
                        try:
                            reference_ct = str(meta.ReferencedImageSequence[0].ReferencedSOPInstanceUID)
                        except:
                            pass
                        try:
                            reference_pl = str(meta.ReferencedRTPlanSequence[0].ReferencedSOPInstanceUID)
                        except:
                            pass
                
                # MRI Tags
                try:
                    tr = float(meta.RepetitionTime)
                except:
                    pass
                try:
                    te = float(meta.EchoTime)
                except:
                    pass
                try:
                    scan_seq = str(meta.ScanningSequence)
                except:
                    pass
                try:
                    tesla = float(meta.MagneticFieldStrength)
                except:
                    pass
                try:
                    elem = str(meta.ImagedNucleus)
                except:
                    pass
                
                # Frame of Reference UIDs
                try:
                    reference_frame = str(meta.FrameOfReferenceUID)
                except:
                    try:
                        reference_frame = str(meta.ReferencedFrameOfReferenceSequence[0].FrameOfReferenceUID)
                    except:
                        reference_frame = ""
        
                try:
                    study_description = str(meta.StudyDescription)
                except:
                    study_description = ""

                try:
                    series_description = str(meta.SeriesDescription)
                except:
                    series_description = ""

                try:
                    subseries = str(meta.AcquisitionNumber)
                except:
                    subseries = "default"

                if patient not in database:
                    database[patient] = {}
                if study not in database[patient]:
                    database[patient][study] = {'description': study_description}
                if series not in database[patient][study]:
                    rel_crawl_path = rel_posix
                    if meta.Modality == 'RTSTRUCT':
                        rel_crawl_path = os.path.join(rel_crawl_path, fname)
                    
                    database[patient][study][series] = {'description': series_description}
                if subseries not in database[patient][study][series]:
                    database[patient][study][series][subseries] = {'instances': {},
                                                                   'instance_uid': instance,
                                                                   'modality': meta.Modality,
                                                                   'reference_ct': reference_ct,
                                                                   'reference_rs': reference_rs,
                                                                   'reference_pl': reference_pl,
                                                                   'reference_frame': reference_frame,
                                                                   'folder': rel_crawl_path,
                                                                   'orientation': orientation,
                                                                   'orientation_type': orientation_type,
                                                                   'repetition_time':tr,
                                                                   'echo_time':te,
                                                                   'scan_sequence': scan_seq,
                                                                   'mag_field_strength': tesla,
                                                                   'imaged_nucleus': elem,
                                                                   'fname': rel_path.as_posix()  # temporary until we switch to json-based loading
                                                                   }
                database[patient][study][series][subseries]['instances'][instance] = rel_path.as_posix()
            except Exception as e:
                print(folder, e)
                pass
    
    return database


def to_df(database_dict):
    df = pd.DataFrame()
    for pat in database_dict:
        for study in database_dict[pat]:
            for series in database_dict[pat][study]:
                if series != 'description':  # skip description key in dict
                    for subseries in database_dict[pat][study][series]:
                        if subseries != 'description':  # skip description key in dict
                            columns = ['patient_ID', 'study', 'study_description', 
                                       'series', 'series_description', 'subseries', 'modality', 
                                       'instances', 'instance_uid', 
                                       'reference_ct', 'reference_rs', 'reference_pl', 'reference_frame', 'folder',
                                       'orientation', 'orientation_type', 'MR_repetition_time', 'MR_echo_time', 
                                       'MR_scan_sequence', 'MR_magnetic_field_strength', 'MR_imaged_nucleus', 'file_path']
                            values = [pat, study, database_dict[pat][study]['description'], 
                                      series, database_dict[pat][study][series]['description'], 
                                      subseries, database_dict[pat][study][series][subseries]['modality'], 
                                      len(database_dict[pat][study][series][subseries]['instances']), database_dict[pat][study][series][subseries]['instance_uid'], 
                                      database_dict[pat][study][series][subseries]['reference_ct'], database_dict[pat][study][series][subseries]['reference_rs'], 
                                      database_dict[pat][study][series][subseries]['reference_pl'], database_dict[pat][study][series][subseries]['reference_frame'], database_dict[pat][study][series][subseries]['folder'],
                                      database_dict[pat][study][series][subseries]['orientation'], database_dict[pat][study][series][subseries]['orientation_type'],
                                      database_dict[pat][study][series][subseries]['repetition_time'], database_dict[pat][study][series][subseries]['echo_time'],
                                      database_dict[pat][study][series][subseries]['scan_sequence'], database_dict[pat][study][series][subseries]['mag_field_strength'], database_dict[pat][study][series][subseries]['imaged_nucleus'],
                                      database_dict[pat][study][series][subseries]['fname']
                                      ]

                            df_add = pd.DataFrame([values], columns=columns)
                            df = pd.concat([df, df_add], ignore_index=True)
    return df


def crawl(top, 
          n_jobs: int = -1):
    # top is the input directory in the argument parser from autotest.py
    database_list = []
    folders = glob.glob(pathlib.Path(top, "*").as_posix())
    
    database_list = Parallel(n_jobs=n_jobs)(delayed(crawl_one)(pathlib.Path(top, folder).as_posix()) for folder in tqdm(folders))

    # convert list to dictionary
    database_dict = {}
    for db in database_list:
        for key in db:
            database_dict[key] = db[key]
    
    # save one level above imaging folders
    parent, dataset  = os.path.split(top)

    parent_imgtools = pathlib.Path(parent, ".imgtools").as_posix()

    if not os.path.exists(parent_imgtools):
        try:
            os.makedirs(parent_imgtools)
        except:
            pass
    
    # save as json
    with open(pathlib.Path(parent_imgtools, f'imgtools_{dataset}.json').as_posix(), 'w') as f:
        json.dump(database_dict, f, indent=4)
    
    # save as dataframe
    df = to_df(database_dict)
    df_path = pathlib.Path(parent_imgtools, f'imgtools_{dataset}.csv').as_posix()
    df.to_csv(df_path)
    
    return database_dict


if __name__ == "__main__":
    parser = ArgumentParser("Dataset DICOM Crawler")
    parser.add_argument("directory",
                        type=str,
                        help="Top-level directory of the dataset.")
    parser.add_argument("--n_jobs",
                        type=int,
                        default=16,
                        help="Number of parallel processes for multiprocessing.")

    args = parser.parse_args()
    db = crawl(args.directory, n_jobs=args.n_jobs)
    print("# patients:", len(db))



import os
import pathlib
import json
import glob
import re
from typing import Optional
from collections import namedtuple
# import copy

import pandas as pd
import SimpleITK as sitk
from pydicom import dcmread

# from joblib import Parallel, delayed
# from tqdm.auto import tqdm

from ..modules import StructureSet, Dose, PET, Scan, Segmentation
from ..utils.crawl import *
from ..utils.dicomutils import *


def read_image(path):
    return sitk.ReadImage(path)


def read_dicom_series(path: str,
                      series_id: Optional[str] = None,
                      recursive: bool = False, 
                      file_names: list = None):
    """Read DICOM series as SimpleITK Image.

    Parameters
    ----------
    path
       Path to directory containing the DICOM series.

    recursive, optional
       Whether to recursively parse the input directory when searching for
       DICOM series,

    series_id, optional
       Specifies the DICOM series to load if multiple series are present in
       the directory. If None and multiple series are present, loads the first
       series found.

    file_names, optional
        If there are multiple acquisitions/"subseries" for an individual series,
        use the provided list of file_names to set the ImageSeriesReader.

    Returns
    -------
    The loaded image.

    """
    reader = sitk.ImageSeriesReader()
    if file_names is None:
        file_names = reader.GetGDCMSeriesFileNames(path,
                                                   seriesID=series_id if series_id else "",
                                                   recursive=recursive)
        # extract the names of the dicom files that are in the path variable, which is a directory
    
    reader.SetFileNames(file_names)
    
    # Configure the reader to load all of the DICOM tags (public+private):
    # By default tags are not loaded (saves time).
    # By default if tags are loaded, the private tags are not loaded.
    # We explicitly configure the reader to load tags, including the
    # private ones.
    reader.MetaDataDictionaryArrayUpdateOn()
    reader.LoadPrivateTagsOn()

    return reader.Execute()

    
def read_dicom_scan(path, series_id=None, recursive: bool=False, file_names=None) -> Scan:
    image = read_dicom_series(path, series_id=series_id, recursive=recursive, file_names=file_names)
    return Scan(image, {})


def read_dicom_rtstruct(path):
    return StructureSet.from_dicom_rtstruct(path)


def read_dicom_rtdose(path):
    return Dose.from_dicom_rtdose(path)


def read_dicom_pet(path, series=None):
    return PET.from_dicom_pet(path, series, "SUV")


def read_dicom_seg(path, meta, series=None):
    seg_img = read_dicom_series(path, series)
    return Segmentation.from_dicom_seg(seg_img, meta)


def read_dicom_auto(path, series=None, file_names=None):
    if path is None:
        return None
    if path.endswith(".dcm"):
        dcms = [path]
    else:
        dcms = glob.glob(pathlib.Path(path, "*.dcm").as_posix())
        
    for dcm in dcms:
        meta = dcmread(dcm)
        if meta.SeriesInstanceUID != series and series is not None:
            continue
        
        modality = meta.Modality
        if modality in ['CT', 'MR']:
            obj = read_dicom_scan(path, series, file_names=file_names)
        elif modality == 'PT':
            obj = read_dicom_pet(path, series)
        elif modality == 'RTSTRUCT':
            obj = read_dicom_rtstruct(dcm)
        elif modality == 'RTDOSE':
            obj = read_dicom_rtdose(dcm)
        elif modality == 'SEG':
            obj = read_dicom_seg(path, meta, series)
        else:
            if len(dcms) == 1:
                print(modality, 'at', dcms[0], 'is NOT implemented yet.')
                raise NotImplementedError
            else:
                print("There were no dicoms in this path.")
                return None
        
        obj.metadata.update(get_modality_metadata(meta, modality))
        return obj


class BaseLoader:
    def __getitem__(self, subject_id):
        raise NotImplementedError

    def __len__(self):
        return len(self.keys())

    def keys(self):
        raise NotImplementedError

    def items(self):
        return ((k, self[k]) for k in self.keys())

    def values(self):
        return (self[k] for k in self.keys())

    def get(self, subject_id, default=None):
        try:
            return self[subject_id]
        except KeyError:
            return default


class ImageTreeLoader(BaseLoader):
    def __init__(self,
                 json_path,
                 csv_path_or_dataframe,
                 col_names=[],
                 study_names=[],
                 series_names=[],
                 subseries_names=[],
                 id_column=None,
                 expand_paths=False,
                 readers=None):

        if readers is None:
            readers = [read_image]  # no mutable defaults https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/

        self.expand_paths = expand_paths
        self.readers = readers
        self.colnames = col_names
        self.studynames = study_names
        self.seriesnames = series_names
        self.subseriesnames = subseries_names

        if isinstance(csv_path_or_dataframe, str):
            if id_column is not None and id_column not in self.colnames:
                self.colnames.append(id_column)
            self.paths = pd.read_csv(csv_path_or_dataframe,
                                     index_col=id_column)
        elif isinstance(csv_path_or_dataframe, pd.DataFrame):
            self.paths = csv_path_or_dataframe
            if id_column:
                self.paths = self.paths.set_index(id_column)
            if len(self.colnames) == 0:
                self.colnames = self.paths.columns
        else:
            raise ValueError(f"Expected a path to csv file or pd.DataFrame, not {type(csv_path_or_dataframe)}.")
        
        if isinstance(json_path, str):
            with open(json_path, 'r') as f:
                self.tree = json.load(f)
        else:
            raise ValueError(f"Expected a path to a json file, not {type(json_path)}.")

        if not isinstance(readers, list):
            readers = [readers] * len(self.colnames)

        self.output_tuple = namedtuple("Output", self.colnames)

    def __getitem__(self, subject_id):
        row = self.paths.loc[subject_id]
        paths = {col: row[col] for col in self.colnames}
        study = {col: row[col] for col in self.studynames}
        series = {col: row[col] for col in self.seriesnames}
        subseries = {col: row[col] for col in self.subseriesnames}
        paths = {k: v if pd.notna(v) else None for k, v in paths.items()}
        
        if self.expand_paths:
            # paths = {col: glob.glob(path)[0] for col, path in paths.items()}
            paths = {col: glob.glob(path)[0] if pd.notna(path) else None for col, path in paths.items()}
        
        for i, (col, path) in enumerate(paths.items()):
            files = self.tree[subject_id][study["study_"+("_").join(col.split("_")[1:])]][series["series_"+("_").join(col.split("_")[1:])]][subseries["subseries_"+("_").join(col.split("_")[1:])]]
            self.readers[i](path, series["series_"+("_").join(col.split("_")[1:])])
        outputs = {col: self.readers[i](path, series["series_"+("_").join(col.split("_")[1:])], file_names=files) for i, (col, path) in enumerate(paths.items())}
        return self.output_tuple(**outputs)

    def keys(self):
        return list(self.paths.index)

    def items(self):
        return ((k, self[k]) for k in self.keys())
    

class ImageCSVLoader(BaseLoader):
    def __init__(self,
                 csv_path_or_dataframe,
                 colnames=[],
                 seriesnames=[],
                 id_column=None,
                 expand_paths=False,
                 readers=None):

        if readers is None:
            readers = [read_image]  # no mutable defaults https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/

        self.expand_paths = expand_paths
        self.readers = readers

        self.colnames = colnames
        self.seriesnames = seriesnames
        if isinstance(csv_path_or_dataframe, str):
            if id_column is not None and id_column not in colnames:
                colnames.append(id_column)
            self.paths = pd.read_csv(csv_path_or_dataframe,
                                     index_col=id_column)
        elif isinstance(csv_path_or_dataframe, pd.DataFrame):
            self.paths = csv_path_or_dataframe
            if id_column:
                self.paths = self.paths.set_index(id_column)
            if len(self.colnames) == 0:
                self.colnames = self.paths.columns
        else:
            raise ValueError(f"Expected a path to csv file or pd.DataFrame, not {type(csv_path_or_dataframe)}.")

        if not isinstance(readers, list):
            readers = [readers] * len(self.colnames)

        self.output_tuple = namedtuple("Output", self.colnames)

    def __getitem__(self, subject_id):
        row = self.paths.loc[subject_id]
        paths = {col: row[col] for col in self.colnames}
        series = {col: row[col] for col in self.seriesnames}
        paths = {k: v if pd.notna(v) else None for k, v in paths.items()}
        if self.expand_paths:
            # paths = {col: glob.glob(path)[0] for col, path in paths.items()}
            paths = {col: glob.glob(path)[0] if pd.notna(path) else None for col, path in paths.items()}
        
        outputs = {col: self.readers[i](path,series["series_"+("_").join(col.split("_")[1:])]) for i, (col, path) in enumerate(paths.items())}
        return self.output_tuple(**outputs)

    def keys(self):
        return list(self.paths.index)

    def items(self):
        return ((k, self[k]) for k in self.keys())


class ImageFileLoader(BaseLoader):
    def __init__(self,
                 root_directory,
                 get_subject_id_from="filename",
                 subdir_path=None,
                 exclude_paths=None,
                 reader=None):

        if exclude_paths is None:
            exclude_paths = []
        if reader is None:
            reader = read_image  # no mutable defaults https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/

        self.root_directory = root_directory
        self.get_subject_id_from = get_subject_id_from
        self.subdir_path = subdir_path
        self.exclude_paths = []
        for path in exclude_paths:
            if not path.startswith(self.root_directory):
                full_paths = glob.glob(pathlib.Path(root_directory, path).as_posix())
                self.exclude_paths.extend(full_paths)
            else:
                full_path = path
                self.exclude_paths.append(full_path)
        self.reader = reader

        self.paths = self._generate_paths()

    def _generate_paths(self):
        paths = {}
        for f in os.scandir(self.root_directory):
            if f.path in self.exclude_paths:
                continue
            subject_dir_path = f.path
            if self.subdir_path:
                full_path = pathlib.Path(subject_dir_path, self.subdir_path).as_posix()
            else:
                full_path = subject_dir_path
            try:
                full_path = glob.glob(full_path)[0]
            except IndexError:
                continue
            if os.path.isdir(full_path):
                full_path = pathlib.Path(full_path, "").as_posix()
            subject_dir_name = os.path.basename(os.path.normpath(subject_dir_path))
            subject_id = self._extract_subject_id_from_path(full_path, subject_dir_name)
            paths[subject_id] = full_path
        return paths

    def _extract_subject_id_from_path(self, full_path, subject_dir_name):
        filename, _ = os.path.splitext(os.path.basename(full_path))
        if isinstance(self.get_subject_id_from, str):
            if self.get_subject_id_from == "filename":
                subject_id = filename
            elif self.get_subject_id_from == "subject_directory":
                subject_id = subject_dir_name
            else:
                subject_id = re.search(self.get_subject_id_from, full_path)[0]
        else:
            return self.get_subject_id_from(full_path, filename, subject_dir_name)
        return subject_id

    def __getitem__(self, subject_id):
        path = self.paths[subject_id]
        return self.reader(path)

    def keys(self):
        return self.paths.keys()


# class CombinedLoader(BaseLoader):
#     def __init__(self, **kwargs):
#         self.loaders = kwargs
#         self.output_tuple = namedtuple("Output", list(self.loaders.keys()))

#     def __getitem__(self, subject_id):
#         outputs = {name: loader[subject_id] for name, loader in self.loaders.items()}
#         return self.output_tuple(**outputs)

#     def keys(self):
#         return set(chain.from_iterable(loader.keys() for loader in self.loaders))

#     def items(self):
#         return ((k, self[k]) for k in self.keys())



import SimpleITK as sitk
import numpy as np


def physical_points_to_idxs(image, points, continuous=False):
    if continuous:
        transform = image.TransformPhysicalPointToContinuousIndex
    else:
        transform = image.TransformPhysicalPointToIndex
    
    vectorized_transform = np.vectorize(lambda x: np.array(transform(x)), signature='(3)->(3)')
    
    # transform indices to ContourSequence/ContourData-wise
    t_points = []
    for slc in points:
        t_points.append(vectorized_transform(slc)[:,::-1])
    return t_points


def idxs_to_physical_points(image, idxs):
    continuous = any([isinstance(i, float) for i in idxs])

    if continuous:
        transform = image.TransformContinuousIndexToPhysicalPoint
    else:
        transform = image.TransformIndexToPhysicalPoint
    vectorized_transform = np.vectorize(lambda x: np.array(transform(x)), signature='(3)->(3)')
    return vectorized_transform(idxs)


def image_to_array(image):
    origin, direction, spacing = image.GetOrigin(), image.GetDirection(), image.GetSpacing()
    array = sitk.GetArrayFromImage(image)
    return array, origin, direction, spacing


def show_image(image, mask=None, ax=None):
    import matplotlib.pyplot as plt
    if ax is None:
        ax = plt.subplots()

    image_array, *_ = image_to_array(image)

    ax.imshow(image_array, cmap="bone", interpolation="bilinear")

    if mask is not None:
        mask_array, *_ = image_to_array(mask)
        mask_array = np.ma.masked_where(mask_array == 0, mask_array)

        ax.imshow(mask_array, cmap="tab20")

    return ax



import os
import pathlib
import ast
from typing import List, Sequence, Tuple
from tqdm import tqdm

import SimpleITK as sitk
import torchio as tio
import pandas as pd

from imgtools.ops import StructureSetToSegmentation, ImageAutoInput, Resample, BaseOp
from imgtools.pipeline import Pipeline
from joblib import Parallel, delayed


class Dataset(tio.SubjectsDataset):
    """
    This class takes in medical dataset in the form of nrrds or directly from the dataset and converts the data into torchio.Subject object, which can be loaded into 
    torchio.SubjectDataset object.
    This class inherits from torchio.SubjectDataset object, which can support transforms and torch.Dataloader.
    Read more about torchio from https://torchio.readthedocs.io/quickstart.html and torchio.SubjectDataset from https://github.com/fepegar/torchio/blob/3e07b78da16d6db4da7193325b3f9cb31fc0911a/torchio/data/dataset.py#L101
    """
    def __init__(self,
                 subjects: Sequence[tio.Subject],
                 path: str) -> List[tio.Subject]:
                
        super().__init__(subjects)
        self.subjects = subjects
        self.path = path
    
    @classmethod
    def load_image(
            cls,
            path:str,
            ignore_multi: bool = True,
            ) -> List[tio.Subject]:
        """
        Based on the given path, passess the processed nrrd files present in the directory and the metadata associated with it and creates a list of Subject instances
        Parameters
            path: Path to the output directory passed to the autopipeline script. The output directory should have all the user mentioned modalities processed and present in their folder. The directory
                  should additionally have dataset.csv which stores all the metadata
        """
        path_metadata = pathlib.Path(path, "dataset.csv").as_posix()
        if not os.path.exists(path_metadata):
            raise ValueError("The specified path has no file name {}".format(path_metadata))
        df_metadata = pd.read_csv(path_metadata,index_col=0)
        
        for col in df_metadata.columns:
            if col.startswith("output_folder"):
                df_metadata[col] = df_metadata[col].apply(lambda x: pathlib.Path(os.path.split(os.path.dirname(path))[0], x).as_posix() if isinstance(x, str) else x)  # input folder joined with the rel path
        
        output_streams = [("_").join(cols.split("_")[2:]) for cols in df_metadata.columns if cols.split("_")[0] == "output"]
        imp_metadata = [cols for cols in df_metadata.columns if cols.split("_")[0] in ("metadata")]
        # Ignores multiple connection to single modality
        if ignore_multi:
            output_streams = [items for items in output_streams if not items.split("_")[-1].isnumeric()]
            imp_metadata = [items for items in imp_metadata if not items.split("_")[-1].isnumeric()]
        
        # Based on the file naming convention
        subject_id_list = list(df_metadata.index)
        subjects = []
        for subject_id in tqdm(subject_id_list):
            temp = {}
            for col in output_streams:
                metadata_name = f"metadata_{col}"
                if 'RTSTRUCT' in col:
                    # print("df in dataset")
                    # print(df_metadata.loc[subject_id])
                    filenames = ast.literal_eval(df_metadata.loc[subject_id]['metadata_RTSTRUCT_CT'])[0]
                    filename = filenames[0]
                else:
                    filename = col
                path_mod = pathlib.Path(path, subject_id, col, f"{filename}.nii.gz").as_posix()
                print(path_mod)
                # All modalities except RTSTRUCT should be of type torchIO.ScalarImage 
                if os.path.exists(path_mod):
                    if col.split("_")[0]!="RTSTRUCT":
                        temp[f"mod_{col}"] = tio.ScalarImage(path_mod)
                    else:
                        path_mods = [pathlib.Path(path, subject_id, col, f"{filename}.nii.gz").as_posix() for filename in filenames]
                        temp[f"mod_{col}"] = tio.LabelMap(path_mods)
                else:
                    temp[f"mod_{col}"] = None
                # For including metadata
                if metadata_name in imp_metadata:
                    # convert string to proper datatype
                    meta = df_metadata.loc[subject_id,metadata_name]
                    if pd.notna(meta):
                        temp[metadata_name] = eval(meta)[0]
                    else:
                        # torch dataloader doesnt accept None type
                        temp[metadata_name] = {}
            subjects.append(tio.Subject(temp))
        return cls(subjects, path)

    @classmethod
    def load_directly(
            cls,
            path:str,
            modalities: str,
            n_jobs: int = -1,
            spacing: Tuple = (1., 1., 0.),
            ignore_multi: bool = True,
            ) -> List[tio.Subject]:
        """
        Based on the given path, imgtools crawls through the directory, forms datagraph and picks the user defined modalities. These paths are processed into sitk.Image.
        This image and the metadata associated with it, creates a list of Subject instances
        Parameters
            path: Path to the directory of the dataset
        """
        input = ImageAutoInput(path, modalities, n_jobs)
        df_metadata = input.df_combined
        output_streams = input.output_streams
        # Ignores multiple connection to single modality
        if ignore_multi:
            output_streams = [items for items in output_streams if not items.split("_")[-1].isnumeric()]
        # Basic operations
        subject_id_list = list(df_metadata.index)
        # basic image processing ops
        resample = Resample(spacing=spacing)
        make_binary_mask = StructureSetToSegmentation(roi_names=[], continuous=False)
        subjects =  Parallel(n_jobs=n_jobs)(delayed(cls.process_one_subject)(input,subject_id,output_streams,resample,make_binary_mask) for subject_id in tqdm(subject_id_list))
        return cls(subjects,path)

    @staticmethod
    def process_one_subject(
            input: Pipeline,
            subject_id: str,
            output_streams: List[str],
            resample: BaseOp,
            make_binary_mask: BaseOp,  
            ) -> tio.Subject:
        """
        Process all modalities for one subject
        Parameters:
            input: ImageAutoInput class which helps in loading the respective DICOMs
            subject_id: subject id of the data
            output_streams: the modalities that are being considered, Note that there can be multiple items of same modality based on their relations with different modalities
            resample: transformation which resamples sitk.Image
            make_binary_mask: transformation useful in making binary mask for rtstructs
        Returns tio.Subject instance for a particular subject id
        """
        temp = {}
        read_results = input(subject_id)
        for i,colname in enumerate(output_streams):
            modality = colname.split("_")[0]
            output_stream = ("_").join([item for item in colname.split("_") if item != "1"])

            if read_results[i] is None:
                temp[f"mod_{colname}"] = None
            elif modality == "CT":
                image = read_results[i]
                if len(image.GetSize()) == 4:
                    assert image.GetSize()[-1] == 1, f"There is more than one volume in this CT file for {subject_id}."
                    extractor = sitk.ExtractImageFilter()
                    extractor.SetSize([*image.GetSize()[:3], 0])
                    extractor.SetIndex([0, 0, 0, 0])    
                    image = extractor.Execute(image)
                image = resample(image)
                temp[f"mod_{colname}"] = tio.ScalarImage.from_sitk(image)
            elif modality == "RTDOSE":
                try:  # For cases with no image present
                    doses = read_results[i].resample_dose(image)
                except:
                    Warning("No CT image present. Returning dose image without resampling")
                    doses = read_results[i]
                temp[f"mod_{colname}"] = tio.ScalarImage.from_sitk(doses)
                temp[f"metadata_{colname}"] = read_results[i].get_metadata()
            elif modality == "PT":
                try:
                    # For cases with no image present
                    pet = read_results[i].resample_pet(image)
                except:
                    Warning("No CT image present. Returning PT/PET image without resampling.")
                    pet = read_results[i]
                temp[f"mod_{colname}"] = tio.ScalarImage.from_sitk(pet)
                temp[f"metadata_{colname}"] = read_results[i].get_metadata()
            elif modality == "RTSTRUCT":
                # For RTSTRUCT, you need image or PT
                structure_set = read_results[i]
                conn_to = output_stream.split("_")[-1]
                # make_binary_mask relative to ct/pet
                if conn_to == "CT":
                    mask = make_binary_mask(structure_set, image)
                elif conn_to == "PT":
                    mask = make_binary_mask(structure_set, pet)  # noqa: F821
                else:
                    raise ValueError("You need to pass a reference CT or PT/PET image to map contours to.")
                temp[f"mod_{colname}"] = tio.LabelMap.from_sitk(mask)
                temp[f"metadata_{colname}"] = structure_set.roi_names
            
        return tio.Subject(temp)
